/*-
 * Copyright 2018 UPLEX Nils Goroll Systemoptimierung
 *
 * All rights reserved.
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <string.h>	// memset for INIT_OBJ
#include <stdlib.h>	// free/calloc

#include <cache/cache.h>
#include <cache/cache_director.h>
#include <vcl.h>
#include <vtim.h>

#include "vcc_all_healthy_if.h"

struct vmod_all_healthy_director {
	unsigned				magic;
#define VMOD_ALL_HEALTHY_DIRECTOR_MAGIC	0x0d8790a3
	VCL_BACKEND				dir;
	VCL_BACKEND				backend;
	int					spcconsider;
	int					nconsider;
	VCL_BACKEND				*consider;
};

static VCL_BACKEND vmod_director_resolve(VCL_BACKEND, struct worker *,
    struct busyobj *);
static VCL_BOOL vmod_director_healthy(VCL_BACKEND, const struct busyobj *,
    VCL_TIME *);

VCL_VOID
vmod_director__init(VRT_CTX,
    struct vmod_all_healthy_director **dp, const char *vcl_name)
{
	struct director *dir;
	struct vmod_all_healthy_director *d;
	const int spc = 4;

	AN(dp);
	AZ(*dp);
	ALLOC_OBJ(d, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);
	if (d == NULL) {
		VRT_fail(ctx, "obj alloc failed");
		return;
	}
	d->consider = calloc(spc, sizeof(VCL_BACKEND));
	if (d->consider == NULL) {
		VRT_fail(ctx, "consider list alloc failed");
		goto fail_consider;
	}
	d->spcconsider = spc;

	ALLOC_OBJ(dir, DIRECTOR_MAGIC);
	if (dir == NULL) {
		VRT_fail(ctx, "director alloc failed");
		goto fail_dir;
	}

	dir->name = "all_healthy";
	REPLACE(dir->vcl_name, vcl_name);
	dir->priv = d;
	dir->healthy = vmod_director_healthy;
	dir->resolve = vmod_director_resolve;
	dir->admin_health = VDI_AH_HEALTHY;
	d->dir = dir;
	*dp = d;
	return;

  fail_dir:
	free(TRUST_ME(d->consider));
  fail_consider:
	FREE_OBJ(d);
	return;
}

VCL_VOID
vmod_director__fini(struct vmod_all_healthy_director **dp) {
	struct vmod_all_healthy_director *d = *dp;

	*dp = NULL;
	if (d == NULL)
		return;
	CHECK_OBJ(d, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);
	free(TRUST_ME(d->dir->vcl_name));
	free(TRUST_ME(d->dir));
	free(TRUST_ME(d->consider));
	FREE_OBJ(d);
}

#define check_init(ctx, name) do {					\
		if (((ctx)->method & VCL_MET_INIT) == 0) {		\
			VRT_fail(ctx, "." #name				\
			    " can only be called from vcl_init {}");	\
			return;						\
		}							\
	} while(0)

VCL_VOID
vmod_director_consider(VRT_CTX,
    struct vmod_all_healthy_director *d, VCL_BACKEND b)
{
	int i;
	void *n;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ(d, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);
	check_init(ctx, consider);

	if (b == NULL) {
		VRT_fail(ctx, "cannot consider a NULL backend");
		return;
	}
	for (i = 0; i < d->nconsider; i++) {
		if (d->consider[i] == b)
			return;
	}
	if (d->spcconsider == d->nconsider) {
		i = d->spcconsider << 1;
		n = realloc(d->consider, i * sizeof(VCL_BACKEND));
		if (n == NULL) {
			VRT_fail(ctx, "growing consider list failed");
			return;
		}
		d->spcconsider = i;
		d->consider = n;
	}
	d->consider[d->nconsider++] = b;
	if (d->backend == NULL)
		d->backend = b;
}

VCL_VOID
vmod_director_set_backend(VRT_CTX,
    struct vmod_all_healthy_director *d, VCL_BACKEND b)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ(d, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);
	check_init(ctx, consider);

	if (b == NULL) {
		VRT_fail(ctx, "cannot set a NULL backend");
		return;
	}

	d->backend = b;
}

VCL_BACKEND
vmod_director_backend(VRT_CTX, struct vmod_all_healthy_director *d)
{
	CHECK_OBJ(d, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);

	(void) ctx;

	return (d->dir);
}

static VCL_BACKEND vmod_director_resolve(VCL_BACKEND b, struct worker *w,
    struct busyobj *bo)
{
	struct vmod_all_healthy_director *d;

	CAST_OBJ_NOTNULL(d, b->priv, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);
	(void) w;
	(void) b;

	return (d->backend);
}

/*
 * we can only approximate the change time for the unhealthy case because we do
 * not have the full record: consider all backends healthy, a goes down, then b,
 * a comes back up: we would need to record the time a went down initially. We
 * approximate by returning the time b went down, so:
 *
 * - unhealthy: earliest change of unhealthy backend
 * - healthy: latest change of healthy backend
 */

static VCL_BOOL
vmod_director_healthy(VCL_BACKEND b, const struct busyobj *bo,
    VCL_TIME *t)
{
	struct vmod_all_healthy_director *d;
	int i;
	VCL_BOOL br, r = 1;
	VCL_TIME bt, tt[2];
	VCL_BACKEND be;

	CAST_OBJ_NOTNULL(d, b->priv, VMOD_ALL_HEALTHY_DIRECTOR_MAGIC);

	if (t) {
		tt[0] = VTIM_real();
		tt[1] = 0.0;
	}

	for (i = 0; i < d->nconsider; i++) {
		be = d->consider[i];
		CHECK_OBJ_NOTNULL(be, DIRECTOR_MAGIC);
		br = !!be->healthy(be, bo, &bt);
		r &= br;
		if (t) {
			if (r && br) {
				if (bt > tt[1])
					tt[1] = bt;
			} else if (! br) {
				if (bt < tt[0])
					tt[0] = bt;
			}
		}
	}

	assert(r == 0 || r == 1);
	if (t)
		*t = tt[r];

	return (r);
}
